//-----------------------------------------------------------------------------------------
//   
// uNikeOS v0.3
// Copyright (C) 2005 by Rst7/CBSIE
//-----------------------------------------------------------------------------------------
// :
// ATMega128   RAM 32K
//  WG12864A
//        
// DB0-DB7 - PA0-PA7
// E       - PC7 (   external ram)
// R/W     - PC6
// CS1     - PC5
// CS2     - PC4
// D/I     - PC3
//  -  4*4 (1607)
//  45DBxxx  SPI
// USART0 - RS485
// USART1 - RS232
//  ENC28J60  SPI
//-----------------------------------------------------------------------------------------
//  RC 8 ,      8.192 
//-----------------------------------------------------------------------------------------
// :   InitMTOS
//-----------------------------------------------------------------------------------------
//  :
// char CreateTask()
//    :
// __task void (void)
// {
//   TASK_HEADER(8,64,10); //  
//                         // RSTACK ( ),  CSTACK ( )
//                         //   
//    
//   CON_TASK_HEADER(8,64,10); //  
//                             // RSTACK ( ),  CSTACK ( )
//                             //   
//   ....... // ,  !!!
// }
//      __  .
//-----------------------------------------------------------------------------------------
// START_MT, STOP_MT -  / 
//-----------------------------------------------------------------------------------------
// void NextTask() -   . 
//-----------------------------------------------------------------------------------------
//  
//      ,  
// void AllocSema(  )
//-----------------------------------------------------------------------------------------
//   (  0)
// void FreeSema(  )
//-----------------------------------------------------------------------------------------
//   2  
// 1-65535   1-65535 
//   1 MAX_FAST_TIMERS
//   1 MAX_SLOW_TIMERS
//-----------------------------------------------------------------------------------------
//   1
// TIMER * AllocFastTimer(SEMA *s) -     
//   1
// TIMER * AllocSlowTimer(SEMA *s) -     
// void SetTimer(TIMER *t, unsigned int tmr) -    tmr /
//     
//-----------------------------------------------------------------------------------------
// ToDo:
// getchar
//  
// Printf/Edit
// TCP/IP  ( SLIP)
//   (45DBxxx)
//-----------------------------------------------------------------------------------------

#include "iom128.h"
#include "inavr.h"
#include "mtos.h"
#include "string.h"

#define TMR_PRESCALLER 3; // 8.192  -   1000 

PTASK current_task;

CON_TASK MAINTASK;

//  1
TIMER FAST_TIMERS[MAX_FAST_TIMERS];
//  1
TIMER SLOW_TIMERS[MAX_SLOW_TIMERS];

//     
char ElapsedRunIn1ms;

//   
//     8.192 
unsigned int ms_in_second;

//   1  2000 .
unsigned long TimeDate;

//   
//   
unsigned long TimeDateLo=0;
//   
unsigned long TimeDateLoInc=0;
//   
char TimeDateSign=1;

//   
char SomeTaskReady;

//-----------------------------------------------------------------------
// 
//-----------------------------------------------------------------------
__z void UpdateTimers(TIMER *t, char c)
{
  unsigned int *p=(unsigned int *)t;
  unsigned int i;
  do
  {
    i=*p;
    if (i)
    {
     i--;
     *p++=i;
     if (!i)
     {
      *((SEMA *)*p)=0;
     }
     p++;
    }
    else p+=2;
  }
  while(--c);
}

//-----------------------------------------------------------------------
//   1 
//   (.  )
//-----------------------------------------------------------------------
#pragma vector=TIMER0_OVF_vect
__interrupt void UpdateSlowTimers(void)
{
  unsigned int i;
  char c;
  i=ms_in_second;
  ms_in_second=0;
  __enable_interrupt();
  c=OSCCAL;
  if (i>1000) c--; else c++;
  OSCCAL=c;
  UpdateTimers(SLOW_TIMERS,MAX_SLOW_TIMERS);
  c=TimeDateSign;
  TimeDateLo+=TimeDateLoInc;
  if (SREG_C) c++;
  TimeDate+=c;
}

//-----------------------------------------------------------------------
//   1m
//-----------------------------------------------------------------------
TIMER * AllocFastTimer(SEMA *s) //     
{
 static char free=MAX_FAST_TIMERS;
 TIMER *t;
 STOP_MT;
 char c=free;
 if (!c)
 {
   START_MT;
   return(0);
 }
 c--;
 free=c;
 t=FAST_TIMERS+c;
 __disable_interrupt();
 t->flag=s;
 __enable_interrupt();
 START_MT;
 return(t);
}

//-----------------------------------------------------------------------
//   1
//-----------------------------------------------------------------------
TIMER * AllocSlowTimer(SEMA *s) //     
{
 static char free=MAX_SLOW_TIMERS;
 TIMER *t;
 STOP_MT;
 char c=free;
 if (!c)
 {
   START_MT;
   return(0);
 }
 c--;
 free=c;
 t=SLOW_TIMERS+c;
 __disable_interrupt();
 t->flag=s;
 *s=1;
 __enable_interrupt();
 START_MT;
 return(t);
}

//-----------------------------------------------------------------------
//   tmr /
//-----------------------------------------------------------------------
__monitor void SetTimer(TIMER *t, unsigned int tmr)
{
 t->tmr=tmr;
}

//-----------------------------------------------------------------------
//  /
//-----------------------------------------------------------------------
__monitor unsigned long GetTimeDate(void)
{
 return(TimeDate);
}

//-----------------------------------------------------------------------
//  /
//-----------------------------------------------------------------------
__monitor void SetTimeDate(unsigned long td)
{
 TimeDate=td;
}

//-----------------------------------------------------------------------
//  . 
//  STOP_MT!!!!
//-----------------------------------------------------------------------
PTASK GetNextTask(PTASK p)
{
 return(p->next);
}
//   GetNextTask,   inline
#pragma optimize=no_inline
//    
#pragma diag_suppress=Pe940
PTASK NextTask(void)
{
 PTASK ct;
 if (!(ct=(PTASK)setjmp(current_task->jb)))
 {
  ct=GetNextTask(current_task);
  longjmp(ct->jb,(int)ct); //  . 
 }
 else
 {
  current_task=ct;
  return(ct);
 }
}
#pragma diag_default=Pe940

//-----------------------------------------------------------------------
//   1 
//-----------------------------------------------------------------------
#pragma vector=TIMER2_OVF_vect
__interrupt void TaskSwitch(void)
{
 char c;
 __enable_interrupt();
 if (c=ElapsedRunIn1ms)
 {
  if (--c)
   ElapsedRunIn1ms=c;
  else
  {
   SomeTaskReady=1; //-  
   ElapsedRunIn1ms=NextTask()->TMR;
  }
 }
 UpdateTimers(FAST_TIMERS,MAX_FAST_TIMERS);
 ms_in_second++;
}

//-----------------------------------------------------------------------
// 
//-----------------------------------------------------------------------
__x void AllocSema(SEMA *ps)
{
 STOP_MT;
 while(*ps) NextTask();
 *ps=1;
 START_MT;
}

//-----------------------------------------------------------------------
// 
//-----------------------------------------------------------------------
__x void FreeSema(SEMA *ps)
{
 STOP_MT;
 *ps=0;
 NextTask(); //. 
 START_MT;
}

//-----------------------------------------------------------------------
// 
//-----------------------------------------------------------------------
int getchar(void)
{
 CON_TASK *con;
 __disable_interrupt();
 con=(void *)current_task;
 __enable_interrupt();
 if (!con->HAVE_CON) return(-1); // 
 if (!con->active) return(-1); //  
 return(0);
}

//-----------------------------------------------------------------------
// 
//-----------------------------------------------------------------------
int putchar(int sym)
{
 char c=sym;
 unsigned int *p;
 CON_TASK *con;
 __disable_interrupt();
 con=(void *)current_task;
 __enable_interrupt();

 if (!con->HAVE_CON) return(-1); // 
 if (con->esc_curpos)
 {
  // .    
  con->cpos=sym-1;
  con->esc_curpos=0;
  return(sym);
 }
 p=con->screen+con->cpos;
 switch(c)
 {
  case 1:
  con->esc_curpos=1;
  break;
  // 
  case 12:
  con->cpos=0;
  memset(con->screen,0,CON_ROWS*CON_COLS*2);
  con->changed=1;
  break;
  // 
  case 13:
  // 
  case 10:
  con->cpos=(con->cpos/CON_ROWS)*CON_ROWS;
  if (c==13) break;
  c=0;
  if (con->cpos>=CON_ROWS*(CON_COLS)) goto L_SCROLL; // . , 
  con->cpos+=20;
  break;
  // 
  default:
  if (con->cpos>CON_ROWS*CON_COLS) return(-1); //    
  if (con->cpos==CON_ROWS*CON_COLS) // 
  {
   c=1;
   goto L_SCROLL;
  }
  if (*p!=sym) goto L_WRCHAR;
  con->cpos++;
  break;
 }
 return(sym);
 L_SCROLL:
 memcpy(con->screen,con->screen+CON_ROWS,CON_ROWS*(CON_COLS-1)*2);
 memset(con->screen+CON_ROWS*(CON_COLS-1),0,CON_ROWS*2);
 if (c)
 {
  con->cpos=CON_ROWS*(CON_COLS-1); p=con->screen+CON_ROWS*(CON_COLS-1);
  L_WRCHAR:
  *p=sym;
  con->cpos++;
 }
 con->changed=1;
 return(sym);
}

//-----------------------------------------------------------------------
//      . 
//-----------------------------------------------------------------------
__x void CreateTask(__task void (tsk(void)))
{
 PTASK ct;
 STOP_MT;
 if (!(ct=(PTASK)setjmp(current_task->jb)))
 {
  __disable_interrupt();
  __indirect_jump_to((unsigned long)tsk);
 }
 else
 {
  __enable_interrupt();
  current_task=ct;
  START_MT;
 }
}

//-----------------------------------------------------------------------
//   
//-----------------------------------------------------------------------
void InsertTask(PTASK t, char tmr, char havecon)
{
 t->TMR=tmr;
 t->HAVE_CON=havecon;
 t->next=current_task->next;
 current_task->next=t;
 current_task=t;
 __enable_interrupt();
 START_MT;
}

//-----------------------------------------------------------------------
//  
//    
//-----------------------------------------------------------------------
__task void ConsoleManager(void)
{
  TASK_HEADER(8,64,1);
  volatile static __no_init unsigned int ca @ 0x7FFE;

  for(;;)
  {
    STOP_MT;
    while(!MAINTASK.changed) NextTask();
    MAINTASK.changed=0;
    // 
    ca=(unsigned int)&MAINTASK.screen;
    START_MT;
  }
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
__task char __low_level_init(void)
{
 extern void *__RSTACK_in_external_ram;
 __require(__RSTACK_in_external_ram);
 MCUCR=0x80; //  
 XMCRB=0x01; // 7   -  
 DDRA=0; //   
 DDRC=0xFF; //   
 PORTA=0;
 PORTC=0;
 return(1);
}

//-----------------------------------------------------------------------
//  
//-----------------------------------------------------------------------
void InitMTOS(void)
{
 __disable_interrupt();
 MAINTASK.HAVE_CON=1; // 
 current_task=MAINTASK.next=(PTASK)(&MAINTASK);
 NextTask();
 TCCR0=5; //  0 (  )
 ASSR=8;
 TCCR2=TMR_PRESCALLER;
 TIMSK_TOIE0=1;
 TIMSK_TOIE2=1;
 START_MT;
 __enable_interrupt();
 CreateTask(ConsoleManager);
}
